/**
 *
 * Openutils testing library (http://www.openmindlab.com/products/testing.html)
 *
 * Copyright(C) 2005-2013, Openmind S.r.l. http://www.openmindonline.it
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * You may obtain a copy of the License at
 *
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package it.openutils.testing;

import it.openutils.testing.dbunit.CleanInsertOperation;
import it.openutils.testing.dbunit.DatasetType;
import it.openutils.testing.dbunit.MsSqlIndentityCleanInsertOperation;
import it.openutils.testing.dbunit.MsSqlIndentityInsertOperation;

import java.io.IOException;
import java.net.URL;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import javax.sql.DataSource;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.dbunit.DatabaseUnitException;
import org.dbunit.database.DatabaseConfig;
import org.dbunit.database.DatabaseConfig.ConfigProperty;
import org.dbunit.database.DatabaseConnection;
import org.dbunit.database.DatabaseSequenceFilter;
import org.dbunit.database.IDatabaseConnection;
import org.dbunit.dataset.DataSetException;
import org.dbunit.dataset.FilteredDataSet;
import org.dbunit.dataset.IDataSet;
import org.dbunit.dataset.datatype.DefaultDataTypeFactory;
import org.dbunit.dataset.datatype.IDataTypeFactory;
import org.dbunit.dataset.filter.ITableFilter;
import org.dbunit.dataset.filter.SequenceTableFilter;
import org.dbunit.ext.hsqldb.HsqldbDataTypeFactory;
import org.dbunit.ext.mssql.MsSqlDataTypeFactory;
import org.dbunit.ext.mysql.MySqlDataTypeFactory;
import org.dbunit.ext.mysql.MySqlMetadataHandler;
import org.dbunit.ext.oracle.Oracle10DataTypeFactory;
import org.dbunit.ext.oracle.OracleDataTypeFactory;
import org.dbunit.operation.DatabaseOperation;
import org.dbunit.operation.DeleteAllOperation;
import org.dbunit.operation.DeleteOperation;
import org.dbunit.operation.InsertOperation;
import org.dbunit.operation.RefreshOperation;
import org.dbunit.operation.TruncateTableOperation;
import org.dbunit.operation.UpdateOperation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;


/**
 * @author fgiust
 * @version $Id: DbUnitTestContext.java 4166 2012-12-23 14:07:23Z fgiust $
 */
public class DbUnitTestContext
{

    /**
     * Truncate dataset cache. This is kept as a static attribute since the creation of the dataset is very expensive
     * and it doesn't change across tests.
     */
    protected static Map<String, IDataSet> truncateDataSetCache = new HashMap<String, IDataSet>();

    /**
     * Logger.
     */
    private static Logger log = LoggerFactory.getLogger(DbUnitTestContext.class);

    /**
     * The test instance.
     */
    private Object testcase;

    /**
     * Spring application context.
     */
    private ApplicationContext applicationContext;

    /**
     * Dataset cache.
     */
    private Map<String, IDataSet> datasetCache = new HashMap<String, IDataSet>();

    /**
     * Instantiates a new DbUnitTestContext
     * @param testcase test instance
     * @param applicationContext Spring application context
     */
    public DbUnitTestContext(Object testcase, ApplicationContext applicationContext)
    {
        this.testcase = testcase;
        this.applicationContext = applicationContext;
    }

    /**
     * Setup the Database before running the test method.
     * @throws Exception Any exception.
     */
    public void setUpDbUnit() throws Exception
    {
        DbUnitExecution singleDbUnitExecution = testcase.getClass().getAnnotation(DbUnitExecution.class);

        DbUnitExecution[] executions = null;
        if (singleDbUnitExecution != null)
        {
            executions = new DbUnitExecution[]{singleDbUnitExecution };
        }
        else
        {
            DbUnitConfiguration dbUnitConfiguration = testcase.getClass().getAnnotation(DbUnitConfiguration.class);
            if (dbUnitConfiguration != null)
            {
                executions = dbUnitConfiguration.dbUnitExecutions();
            }
        }
        if (executions != null)
        {
            for (DbUnitExecution dbUnitExecution : executions)
            {
                String[] datasets = dbUnitExecution.datasets();
                String dataSourceName = dbUnitExecution.dataSource();
                String schema = dbUnitExecution.schema();
                if (StringUtils.isEmpty(schema))
                {
                    schema = null;
                }

                IDatabaseConnection connection = new DatabaseConnection(
                    getDatasource(dataSourceName).getConnection(),
                    schema);

                try
                {
                    DatabaseConfig config = connection.getConfig();

                    DatabaseMetaData metaData = connection.getConnection().getMetaData();

                    String dbtype = metaData.getDatabaseProductName();
                    int dbVersion = metaData.getDatabaseMajorVersion();

                    Class< ? extends IDataTypeFactory> dataTypeFactory = dbUnitExecution.dataTypeFactory();

                    // autoset properties
                    if (StringUtils.equalsIgnoreCase(dbtype, "Microsoft SQL Server"))
                    {
                        setupProperty(
                            config,
                            DatabaseConfig.PROPERTY_DATATYPE_FACTORY,
                            MsSqlDataTypeFactory.class.getName());
                    }
                    else if (StringUtils.equalsIgnoreCase(dbtype, "mysql"))
                    {
                        setupProperty(
                            config,
                            DatabaseConfig.PROPERTY_METADATA_HANDLER,
                            MySqlMetadataHandler.class.getName());
                        setupProperty(
                            config,
                            DatabaseConfig.PROPERTY_DATATYPE_FACTORY,
                            MySqlDataTypeFactory.class.getName());
                    }
                    else if (StringUtils.equalsIgnoreCase(dbtype, "mysql"))
                    {
                        setupProperty(
                            config,
                            DatabaseConfig.PROPERTY_METADATA_HANDLER,
                            MySqlMetadataHandler.class.getName());
                        setupProperty(
                            config,
                            DatabaseConfig.PROPERTY_DATATYPE_FACTORY,
                            MySqlDataTypeFactory.class.getName());
                    }
                    else if (StringUtils.equalsIgnoreCase(dbtype, "hsqldb"))
                    {
                        setupProperty(
                            config,
                            DatabaseConfig.PROPERTY_DATATYPE_FACTORY,
                            HsqldbDataTypeFactory.class.getName());
                    }
                    else if (StringUtils.equalsIgnoreCase(dbtype, "Oracle"))
                    {
                        if (dbVersion >= 10)
                        {
                            setupProperty(
                                config,
                                DatabaseConfig.PROPERTY_DATATYPE_FACTORY,
                                Oracle10DataTypeFactory.class.getName());
                        }
                        else
                        {
                            setupProperty(
                                config,
                                DatabaseConfig.PROPERTY_DATATYPE_FACTORY,
                                OracleDataTypeFactory.class.getName());
                        }
                    }

                    // if the user configured a specialized dataTypeFactory just use it and override our defaults
                    if (!dataTypeFactory.equals(DefaultDataTypeFactory.class))
                    {
                        config.setProperty(DatabaseConfig.PROPERTY_DATATYPE_FACTORY, dataTypeFactory.newInstance());
                    }

                    for (DbUnitFeature feature : dbUnitExecution.features())
                    {
                        config.setProperty(feature.key(), feature.value());
                    }

                    for (DbUnitProperty property : dbUnitExecution.properties())
                    {
                        setupProperty(config, property);
                    }

                    ITableFilter tableFilter = new RegExpTableFilter(dbUnitExecution.excludedTables());

                    if (dbUnitExecution.truncateAll())
                    {
                        truncateAll(
                            connection,
                            tableFilter,
                            dataSourceName,
                            getDatabaseOperation(dbUnitExecution.truncateOperation()));
                    }

                    DatabaseOperation dbOperation = getDatabaseOperation(dbUnitExecution.insertOperation());

                    // default
                    if (InsertOperation.class.equals(dbOperation.getClass()))
                    {
                        if ("Microsoft SQL Server".equals(dbtype))
                        {
                            dbOperation = new MsSqlIndentityInsertOperation();
                        }
                    }
                    else if (CleanInsertOperation.class.equals(dbOperation.getClass()))
                    {
                        if ("Microsoft SQL Server".equals(dbtype))
                        {
                            dbOperation = new MsSqlIndentityCleanInsertOperation();
                        }
                    }
                    for (String datasetFile : datasets)
                    {
                        importDataSet(
                            createDataset(datasetFile, dbUnitExecution.datasetType()),
                            connection,
                            tableFilter,
                            dataSourceName,
                            dbOperation);
                    }
                }
                finally
                {
                    connection.close();
                }
            }
        }
    }

    /**
     * Setup a dbunit property, handling type conversion if needed.
     * @param config
     * @param property
     */
    private void setupProperty(DatabaseConfig config, DbUnitProperty property)
    {
        setupProperty(config, property.key(), property.value());
    }

    /**
     * @param config
     * @param configProperty
     * @param propertyString
     */
    private void setupProperty(DatabaseConfig config, String propertyKey, String propertyString)
    {
        for (ConfigProperty configProperty : DatabaseConfig.ALL_PROPERTIES)
        {
            if (propertyKey.equals(configProperty.getProperty()))
            {
                Class< ? > propertyType = configProperty.getPropertyType();

                Object propertyValue = propertyString;

                if (Integer.class.equals(propertyType))
                {
                    propertyValue = NumberUtils.toInt(propertyString);
                }
                else
                {
                    // a class name?
                    try
                    {
                        Class< ? > classProperty = Class.forName(propertyString);
                        propertyValue = classProperty.newInstance();
                    }
                    catch (Throwable e)
                    {
                        // no, not a class name, ignore.
                        log.debug(
                            "Got a {} while trying to convert the {} property",
                            e.getClass().getName(),
                            configProperty);
                    }
                }

                config.setProperty(configProperty.getProperty(), propertyValue);
                return;
            }
        }

        log.warn("Unknown property {}, skipping", propertyKey);
    }

    /**
     * Instantiates the givec Database Operation. Standard operations in the <code>org.dbunit.operation</code> package
     * have protected constructors and needs to be accessed using the public static fields in
     * <code>org.dbunit.operation.DatabaseOperation</code>.
     * @param dboperationClass db operation class
     * @return db operation instance
     * @throws InstantiationException if the given db operation class cannot be instantiated
     * @throws IllegalAccessException if the given db operation class cannot be instantiated
     */
    private DatabaseOperation getDatabaseOperation(Class< ? extends DatabaseOperation> dboperationClass)
        throws InstantiationException, IllegalAccessException
    {
        if (UpdateOperation.class.equals(dboperationClass))
        {
            return DatabaseOperation.UPDATE;
        }
        else if (InsertOperation.class.equals(dboperationClass))
        {
            return DatabaseOperation.INSERT;
        }
        else if (RefreshOperation.class.equals(dboperationClass))
        {
            return DatabaseOperation.REFRESH;
        }
        else if (DeleteOperation.class.equals(dboperationClass))
        {
            return DatabaseOperation.DELETE;
        }
        else if (DeleteAllOperation.class.equals(dboperationClass))
        {
            return DatabaseOperation.DELETE_ALL;
        }
        else if (TruncateTableOperation.class.equals(dboperationClass))
        {
            return DatabaseOperation.TRUNCATE_TABLE;
        }
        else if (DeleteOperation.class.equals(dboperationClass))
        {
            return DatabaseOperation.UPDATE;
        }

        return dboperationClass.newInstance();

    }

    /**
     * Creates a dataset instance by fetching a file from the classpath
     * @param datasetFile name of the file, will be loaded from the classpath
     * @param datasetType Dataset Type
     * @return IDataSet instance
     * @throws IOException
     * @throws DataSetException
     */
    private IDataSet createDataset(String datasetFile, DatasetType datasetType) throws IOException, DataSetException
    {
        IDataSet dataSet = datasetCache.get(datasetFile);
        if (dataSet == null)
        {
            URL datasetUrl = getClass().getResource(datasetFile);
            if (datasetUrl == null)
            {
                throw new IllegalArgumentException("Dataset " + datasetFile + " not found");
            }
            dataSet = DbUnitUtils.loadDataset(datasetUrl, datasetType);
            datasetCache.put(datasetFile, dataSet);
        }
        return dataSet;
    }

    private void importDataSet(IDataSet dataSet, IDatabaseConnection connection, ITableFilter tableFilter,
        String dataSourceName, DatabaseOperation databaseOperation) throws SQLException, DataSetException,
        DatabaseUnitException
    {

        if (dataSet == null)
        {
            throw new IllegalArgumentException("dataSet is null");
        }

        IDataSet orderedDataset = new FilteredDataSet(tableFilter, dataSet);
        if (log.isDebugEnabled())
        {
            log.debug("Tables: {}", ArrayUtils.toString(orderedDataset.getTableNames()));
        }

        // if a sorted dataset is available, use table sequence for sorting
        IDataSet truncateDataSet = getTruncateDataset(dataSourceName, connection.getSchema());
        if (truncateDataSet != null)
        {
            ITableFilter filter = new SequenceTableFilter(truncateDataSet.getTableNames());
            orderedDataset = new FilteredDataSet(filter, dataSet);
        }

        if (dataSet != null)
        {
            databaseOperation.execute(connection, orderedDataset);
        }

    }

    /**
     * @param connection
     * @param tableFilter
     * @throws SQLException
     * @throws DataSetException
     * @throws DatabaseUnitException
     */
    private void truncateAll(IDatabaseConnection connection, ITableFilter tableFilter, String dataSourceName,
        DatabaseOperation databaseOperation) throws SQLException, DataSetException, DatabaseUnitException
    {
        IDataSet truncateDataSet = getTruncateDataset(dataSourceName, connection.getSchema());
        if (truncateDataSet == null)
        {
            log.debug("Generating sorted dataset for initial cleanup");
            IDataSet unsortedDataSet = connection.createDataSet();

            if (log.isDebugEnabled())
            {
                log.debug("Unfiltered truncateDataSet: {}", ArrayUtils.toString(unsortedDataSet.getTableNames()));
            }

            // excluded unwanted tables
            unsortedDataSet = new FilteredDataSet(tableFilter, unsortedDataSet);

            // sort tables
            ITableFilter sortingFilter = new DatabaseSequenceFilter(connection, unsortedDataSet.getTableNames());
            truncateDataSet = new FilteredDataSet(sortingFilter, unsortedDataSet);

            storeTruncateDataset(dataSourceName, connection.getSchema(), truncateDataSet);
            log.debug("Sorted dataset generated");
        }

        if (truncateDataSet != null)
        {
            if (log.isDebugEnabled())
            {
                log.debug("Tables truncateDataSet: {}", ArrayUtils.toString(truncateDataSet.getTableNames()));
            }
            databaseOperation.execute(connection, truncateDataSet);
        }
    }

    private IDataSet getTruncateDataset(String datasourceName, String schema)
    {
        return truncateDataSetCache.get(datasourceName + "_" + StringUtils.defaultString(schema));
    }

    private IDataSet storeTruncateDataset(String datasourceName, String schema, IDataSet dataset)
    {
        return truncateDataSetCache.put(datasourceName + "_" + StringUtils.defaultString(schema), dataset);
    }

    /**
     * Loads a named datasource from the spring context.
     * @param name datasource name.
     * @return Datasource instance
     */
    protected DataSource getDatasource(String name)
    {
        if (StringUtils.isEmpty(name))
        {
            Map<String, DataSource> dsMap = applicationContext.getBeansOfType(DataSource.class);
            if (dsMap == null || dsMap.size() != 1)
            {
                throw new RuntimeException(
                    "Unable to find a datasource in spring applicationContext, please specify the datasource bean name "
                        + "using the \"datasource\" attribute of @DbUnitConfiguration");
            }
            return dsMap.values().iterator().next();
        }
        return (DataSource) applicationContext.getBean(name);
    }

}
