/**
 *
 * Openutils testing library (http://www.openmindlab.com/products/testing.html)
 *
 * Copyright(C) 2005-2013, Openmind S.r.l. http://www.openmindonline.it
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * You may obtain a copy of the License at
 *
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package it.openutils.testing;

import it.openutils.testing.dbunit.DatasetType;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.sql.SQLException;

import org.apache.commons.lang3.StringUtils;
import org.dbunit.database.IDatabaseConnection;
import org.dbunit.dataset.DataSetException;
import org.dbunit.dataset.FilteredDataSet;
import org.dbunit.dataset.IDataSet;
import org.dbunit.dataset.excel.XlsDataSet;
import org.dbunit.dataset.filter.ITableFilter;
import org.dbunit.dataset.xml.FlatXmlDataSet;
import org.dbunit.dataset.xml.FlatXmlDataSetBuilder;
import org.dbunit.dataset.xml.XmlDataSet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * @author fgiust
 * @version $Id: DbUnitUtils.java 4166 2012-12-23 14:07:23Z fgiust $
 */
public final class DbUnitUtils
{

    /**
     * Logger.
     */
    private static Logger log = LoggerFactory.getLogger(DbUnitUtils.class);

    /**
     * Do not instantiate.
     */
    private DbUnitUtils()
    {
        // do not instantiate
    }

    /**
     * Exports a list of table to an xml/xsl file
     * @param file The file name we save the dataset to
     * @param connection connection to use
     * @param tableNames optional list of table names
     * @throws IOException An IO Exception.
     * @throws DataSetException A dataset exception.
     * @throws SQLException A SQL Exception.
     * @deprecated use {@link #exportTablesToDir(IDatabaseConnection, File, DatasetType, ITableFilter)}
     */
    @Deprecated
    public static void exportDataset(IDatabaseConnection connection, String[] tableNames, File file)
        throws SQLException, IOException, DataSetException
    {
        exportDataset(connection, tableNames, file, DatasetType.XML);
    }

    /**
     * Exports a list of table to an xml/xsl file
     * @param file The file name we save the dataset to
     * @param connection connection to use
     * @param tableNames optional list of table names
     * @param datasetType dataset Type
     * @throws IOException An IO Exception.
     * @throws DataSetException A dataset exception.
     * @throws SQLException A SQL Exception.
     */
    public static void exportDataset(IDatabaseConnection connection, String[] tableNames, File file,
        DatasetType datasetType) throws SQLException, IOException, DataSetException
    {
        OutputStream fos = null;

        try
        {
            IDataSet fullDataSet = connection.createDataSet();

            if (tableNames != null && tableNames.length > 0)
            {
                fullDataSet = new FilteredDataSet(tableNames, fullDataSet);
            }

            fos = new FileOutputStream(file);
            if (datasetType == DatasetType.EXCEL
                || StringUtils.equalsIgnoreCase(StringUtils.substringAfterLast(file.getName(), "."), "xls"))
            {
                XlsDataSet.write(fullDataSet, fos);
            }
            else if (datasetType == DatasetType.XML)
            {
                XmlDataSet.write(fullDataSet, fos);
            }
            else if (datasetType == DatasetType.FLATXML)
            {
                FlatXmlDataSet.write(fullDataSet, fos);
            }
        }
        finally
        {
            fos.close();
        }

        log.debug("Dataset exported at {}", file.getAbsolutePath());
    }

    /**
     * Exports a full db saving a dataset for each table.
     * @param connection dbunit connection
     * @param directory export directory
     * @param xls if <code>true</code> export will be done using excel datasets
     * @param tableFilter optional table filter
     * @throws SQLException sql exception
     * @throws DataSetException dbunit exception
     * @throws IOException IO exception
     * @deprecated use {@link #exportTablesToDir(IDatabaseConnection, File, DatasetType, ITableFilter)}
     */
    @Deprecated
    public void exportTablesToDir(IDatabaseConnection connection, File directory, boolean xls, ITableFilter tableFilter)
        throws SQLException, DataSetException, IOException
    {
        exportTablesToDir(connection, directory, xls ? DatasetType.EXCEL : DatasetType.XML, tableFilter);
    }

    /**
     * Exports a full db saving a dataset for each table.
     * @param connection dbunit connection
     * @param directory export directory
     * @param datasetType dbunit dataset type
     * @param tableFilter optional table filter
     * @throws SQLException sql exception
     * @throws DataSetException dbunit exception
     * @throws IOException IO exception
     */
    public void exportTablesToDir(IDatabaseConnection connection, File directory, DatasetType datasetType,
        ITableFilter tableFilter) throws SQLException, DataSetException, IOException

    {
        directory.mkdirs();

        IDataSet dataset = connection.createDataSet();
        if (tableFilter != null)
        {
            dataset = new FilteredDataSet(tableFilter, dataset);
        }

        String[] tablenames = dataset.getTableNames();

        for (String table : tablenames)
        {
            File exportFile = new File(directory, table + (datasetType == DatasetType.EXCEL ? ".xls" : ".xml"));

            log.info("Exporting {}", table);
            DbUnitUtils.exportDataset(connection, new String[]{table }, exportFile, datasetType);
        }
    }

    /**
     * @param url dataset url (can be an xml or excel file)
     * @return loaded dataset
     * @throws IOException error reading the file
     * @throws DataSetException error loading the dataset
     * @deprecated use {@link #loadDataset(URL, DatasetType)}
     */
    @Deprecated
    public static IDataSet loadDataset(URL url) throws IOException, DataSetException
    {
        if (url == null)
        {
            throw new IllegalArgumentException("URL is null");
        }

        return loadDataset(url, url.getFile().endsWith(".xls") ? DatasetType.EXCEL : DatasetType.XML);
    }

    /**
     * @param url dataset url (can be an xml or excel file)
     * @param datasetType dataset Type
     * @return loaded dataset
     * @throws IOException error reading the file
     * @throws DataSetException error loading the dataset
     */
    public static IDataSet loadDataset(URL url, DatasetType datasetType) throws IOException, DataSetException
    {
        if (url == null)
        {
            throw new IllegalArgumentException("URL is null");
        }
        InputStream is = url.openStream();
        IDataSet dataSet = null;

        if (is != null)
        {
            log.debug("loading dataset {}", url.getFile());

            if (datasetType == DatasetType.EXCEL || (url.getFile().endsWith(".xls")))
            {
                dataSet = new XlsDataSet(is);
            }
            else if (datasetType == DatasetType.FLATXML)
            {
                dataSet = new FlatXmlDataSetBuilder().build(is);
            }
            else if (datasetType == DatasetType.XML)
            {
                dataSet = new XmlDataSet(is);
            }

            is.close();
        }
        return dataSet;
    }

}
