/**
 *
 * Openutils testing library (http://www.openmindlab.com/products/testing.html)
 *
 * Copyright(C) 2005-2013, Openmind S.r.l. http://www.openmindonline.it
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * You may obtain a copy of the License at
 *
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package it.openutils.testing.collections;

import it.openutils.testing.AssertFormatter;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;


/**
 * @author gcatania
 */
public class CollectionAssert extends AssertFormatter
{

    protected CollectionAssert()
    {
    }

    /**
     * Asserts that a collection is null
     * @param c the collection
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertNull(Collection< ? > c, String message) throws AssertionError
    {
        if (c != null)
        {
            throw assertionFailed(message, "Expected a null {0}, found {1} values instead.", typeName(c), c.size());
        }
    }

    /**
     * Asserts that a collection is null
     * @param c the collection
     * @throws AssertionError if assertion fails
     */
    public static void assertNull(Collection< ? > c) throws AssertionError
    {
        assertNull(c, null);
    }

    /**
     * Asserts that a collection is not null
     * @param c the collection
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertNotNull(Collection< ? > c, String message) throws AssertionError
    {
        if (c == null)
        {
            throw assertionFailed(message, "{0} is null.", capitalizedTypeName(c));
        }
    }

    /**
     * Asserts that a collection is not null
     * @param c the collection
     * @throws AssertionError if assertion fails
     */
    public static void assertNotNull(Collection< ? > c) throws AssertionError
    {
        assertNotNull(c, null);
    }

    /**
     * Asserts that a collection is empty. Fails when the collection is null.
     * @param c the collection
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertEmpty(Collection< ? > c, String message) throws AssertionError
    {
        assertNotNull(c, message);
        if (!c.isEmpty())
        {
            throw assertionFailed(message, "Expected an empty {0}, found {1} values instead.", typeName(c), c.size());
        }
    }

    /**
     * Asserts that a collection is empty. Fails when the collection is null.
     * @param c the collection
     * @throws AssertionError if assertion fails
     */
    public static void assertEmpty(Collection< ? > c) throws AssertionError
    {
        assertEmpty(c, null);
    }

    /**
     * Asserts that a collection has the expected size. Fails if the input collection is null.
     * @param actual the actual collection
     * @param expectedSize the expected size
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertSizeEquals(Collection< ? > actual, int expectedSize, String message) throws AssertionError
    {
        assertNotNull(actual, message);
        if (expectedSize == 0)
        {
            assertEmpty(actual, message);
        }
        else
        {
            int actualSize = actual.size();
            if (actualSize != expectedSize)
            {
                throw assertionFailed(
                    message,
                    "{0} has wrong size: expected {1}, found {2} instead.",
                    capitalizedTypeName(actual),
                    expectedSize,
                    actualSize);
            }
        }
    }

    /**
     * Asserts that a collection has the expected size. Fails if the input collection is null.
     * @param actual the actual collection
     * @param expectedSize the expected size
     * @throws AssertionError if assertion fails
     */
    public static void assertSizeEquals(Collection< ? > actual, int expectedSize) throws AssertionError
    {
        assertSizeEquals(actual, expectedSize, null);
    }

    /**
     * Asserts that two collections are equal, regardless of element order. Allows multiple instances of the same
     * element in any of the collections, and checks that they occur the same number of times in both collections. If
     * you need to assert list equality, use {@code ListAssert} instead. If you need to assert set equality, use {@code
     * SetAssert}.
     * @param actual the actual collection
     * @param expected the expected collection
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertEquals(Collection< ? > actual, Collection< ? > expected, String message)
        throws AssertionError
    {
        if (expected == null)
        {
            assertNull(actual, message);
        }
        else if (expected.isEmpty())
        {
            assertEmpty(actual, message);
        }
        else
        {
            assertSizeEquals(actual, expected.size(), message);

            // handle collections that allow multiple instances of the same element
            Set<Object> allKeys = new HashSet<Object>(expected.size() + actual.size(), 1f);
            allKeys.addAll(actual);
            allKeys.addAll(expected);
            Map<Object, Integer> expectedCounts = getObjectCounts(expected);
            Map<Object, Integer> actualCounts = getObjectCounts(actual);

            for (Object key : allKeys)
            {
                Integer expectedCount = expectedCounts.get(key);
                Integer actualCount = actualCounts.get(key);
                if (expectedCount == null)
                {
                    throw assertionFailed(message, "Element {0} not expected in {1}.", key, typeName(actual));
                }
                if (actualCount == null)
                {
                    throw assertionFailed(message, "Element {0} not found in {1}.", key, typeName(actual));
                }
                if (actualCount.intValue() != expectedCount.intValue())
                {
                    throw assertionFailed(
                        message,
                        "Element {0} expected {1} times, but found {2} times in {3}.",
                        key,
                        expectedCount,
                        actualCount,
                        typeName(actual));
                }
            }
        }
    }

    /**
     * Asserts that two collections are equal, regardless of element order. Allows multiple instances of the same
     * element in any of the collections, and checks that they occur the same number of times in both collections. If
     * you need to assert list equality, use {@code ListAssert} instead. If you need to assert set equality, use {@code
     * SetAssert}.
     * @param actual the actual collection
     * @param expected the expected collection
     * @throws AssertionError if assertion fails
     */
    public static void assertEquals(Collection< ? > actual, Collection< ? > expected) throws AssertionError
    {
        assertEquals(actual, expected, null);
    }

    /**
     * Asserts that a collection is sorted.
     * @param <T> the type of elements in the collection, must be comparable
     * @param c the collection
     * @param strict if true and two elements {@code e1} and {@code e2} such that {@code e1.compareTo(e2) == 0} are
     * found, the assertion will fail
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     * @throws NullPointerException if any element in the input collection is null
     */
    public static <T extends Comparable<T>> void assertSorted(Collection<T> c, boolean strict, String message)
        throws AssertionError, NullPointerException
    {
        assertNotNull(c, message);
        if (c.size() < 2)
        {
            return;
        }
        Iterator<T> i = c.iterator();
        T previous = i.next();
        do
        {
            T next = i.next();
            int compare = next.compareTo(previous);
            if (compare < 0)
            {
                throw assertionFailed(
                    message,
                    "{0} is not sorted: element {1} is greater than element {2} but occurs first.",
                    capitalizedTypeName(c),
                    previous,
                    next);
            }
            if (strict && compare == 0)
            {
                throw assertionFailed(
                    message,
                    "{0} is not strictly sorted: element {1} compares equally to element {2}.",
                    capitalizedTypeName(c),
                    previous,
                    next);
            }
            previous = next;
        }
        while (i.hasNext());

    }

    /**
     * Asserts that a collection is sorted.
     * @param <T> the type of elements in the collection, must be comparable
     * @param c the collection
     * @param strict if true and two elements {@code e1} and {@code e2} such that {@code e1.compareTo(e2) == 0} are
     * found, the assertion will fail
     * @throws AssertionError if assertion fails
     * @throws NullPointerException if any element in the input collection is null
     */
    public static <T extends Comparable<T>> void assertSorted(Collection<T> c, boolean strict) throws AssertionError,
        NullPointerException
    {
        assertSorted(c, strict, null);
    }

    /**
     * Asserts that a collection is loosely sorted, i.e that for every two subsequent elements {@code e1} and {@code e2}
     * in the collection, {@code e1.compareTo(e2) >=0} is true.
     * @param <T> the type of elements in the collection, must be comparable
     * @param c the collection
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     * @throws NullPointerException if any element in the input collection is null
     */
    public static <T extends Comparable<T>> void assertSorted(Collection<T> c, String message) throws AssertionError,
        NullPointerException
    {
        assertSorted(c, false, message);
    }

    /**
     * Asserts that a collection is loosely sorted, i.e that for every two subsequent elements {@code e1} and {@code e2}
     * in the collection, {@code e1.compareTo(e2) >=0} is true.
     * @param <T> the type of elements in the collection, must be comparable
     * @param c the collection
     * @throws AssertionError if assertion fails
     * @throws NullPointerException if any element in the input collection is null
     */
    public static <T extends Comparable<T>> void assertSorted(Collection<T> c) throws AssertionError,
        NullPointerException
    {
        assertSorted(c, false, null);
    }

    /**
     * Asserts that a collection is sorted.
     * @param <T> the type of elements in the collection
     * @param c the collection
     * @param comparator defines an ordering on the collection's element
     * @param strict if true and two elements {@code e1} and {@code e2} such that {@code comparator.compare(e1, e2) ==
     * 0} are found, the assertion will fail
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static <T> void assertSorted(Collection<T> c, Comparator<T> comparator, boolean strict, String message)
        throws AssertionError
    {
        assertNotNull(c, message);
        if (c.size() < 2)
        {
            return;
        }
        Iterator<T> i = c.iterator();
        T previous = i.next();
        do
        {
            T next = i.next();
            int compare = comparator.compare(previous, next);
            if (compare < 0)
            {
                throw assertionFailed(
                    message,
                    "{0} is not sorted according to {3}: element {1} is greater than element {2} but occurs first.",
                    capitalizedTypeName(c),
                    previous,
                    next,
                    comparator);
            }
            if (strict && compare == 0)
            {
                throw assertionFailed(
                    message,
                    "{0} is not strictly sorted according to {3}: element {1} compares equally to element {2}.",
                    capitalizedTypeName(c),
                    previous,
                    next,
                    comparator);
            }
            previous = next;
        }
        while (i.hasNext());

    }

    /**
     * Asserts that a collection is sorted.
     * @param <T> the type of elements in the collection
     * @param c the collection
     * @param comparator defines an ordering on the collection's element
     * @param strict if true and two elements {@code e1} and {@code e2} such that {@code comparator.compare(e1, e2) ==
     * 0} are found, the assertion will fail
     * @throws AssertionError if assertion fails
     */
    public static <T> void assertSorted(Collection<T> c, Comparator<T> comparator, boolean strict)
        throws AssertionError
    {
        assertSorted(c, comparator, strict, null);
    }

    /**
     * Asserts that a collection is loosely sorted, i.e. that for every two subsequent elements {@code e1} and {@code
     * e2} in the collection, {@code comparator.compare(e1, e2) >= 0} is true.
     * @param <T> the type of elements in the collection
     * @param c the collection
     * @param comparator defines an ordering on the collection's element
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static <T> void assertSorted(Collection<T> c, Comparator<T> comparator, String message)
        throws AssertionError
    {
        assertSorted(c, comparator, false, message);
    }

    /**
     * Asserts that a collection is loosely sorted, i.e. that for every two subsequent elements {@code e1} and {@code
     * e2} in the collection, {@code comparator.compare(e1, e2) >= 0} is true.
     * @param <T> the type of elements in the collection
     * @param c the collection
     * @param comparator defines an ordering on the collection's element
     * @throws AssertionError if assertion fails
     */
    public static <T> void assertSorted(Collection<T> c, Comparator<T> comparator) throws AssertionError
    {
        assertSorted(c, comparator, false, null);
    }

    /**
     * internal utility method to retrieve the occurrences of distinct objects in a collection.
     * @param c the collection
     * @return a map between objects and their occurrences in the collection
     */
    protected static Map<Object, Integer> getObjectCounts(Collection< ? > c)
    {
        Map<Object, Integer> count = new HashMap<Object, Integer>(c.size(), 1f);
        for (Object o : c)
        {
            Integer currCount = count.get(o);
            if (currCount == null)
            {
                currCount = 1;
            }
            else
            {
                currCount++;
            }
            count.put(o, currCount);
        }
        return count;
    }

}
