/**
 *
 * Openutils testing library (http://www.openmindlab.com/products/testing.html)
 *
 * Copyright(C) 2005-2013, Openmind S.r.l. http://www.openmindonline.it
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * You may obtain a copy of the License at
 *
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package it.openutils.testing.collections;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * Provides short-named methods to build collections. May or may not make your test code unreadable. Example use:
 * 
 * <pre>
 * package com.acme.test;
 * 
 * import static it.openutils.testing.collections.CollectionBuilder.*;
 * 
 * public class FooTest {
 * 
 *     private final Collection coll = Arrays.asList("dum", "de", "dum");
 * 
 *     public void testBar() {
 *         CollectionAssert.assertEquals(coll, c("dum", "dum", "de"));
 *     }
 * }
 * </pre>
 * @author gcatania
 */
public final class CollectionBuilder
{

    private CollectionBuilder()
    {
    };

    /**
     * shorthand method to build a set (with predictable iteration order)
     * @param <T> the objects'class
     * @param args the set arguments
     * @return an hash set with the input arguments
     */
    public static <T> Set<T> s(T... args)
    {
        return new LinkedHashSet<T>(Arrays.asList(args));
    }

    /**
     * shorthand method to build a collection (with predictable iteration order)
     * @param <T> the objects'class
     * @param args the set arguments
     * @return a collection with the input arguments
     */
    public static <T> Collection<T> c(T... args)
    {
        return l(args);
    }

    /**
     * shorthand method to build a list
     * @param <T> the objects'class
     * @param args the set arguments
     * @return a list with the input arguments
     */
    public static <T> List<T> l(T... args)
    {
        return Arrays.asList(args);
    }

    /**
     * shorthand method to build a map
     * @param <K> the key class
     * @param <V> the value class
     * @param keys the keys
     * @param values the values
     * @return a map between the respective input keys and values, as returned by the input collections' iterator (note
     * that this method may yield unpredictable results if the input collections do not have predictable iteration
     * order)
     */
    public static <K, V> Map<K, V> m(Collection<K> keys, Collection<V> values)
    {
        Map<K, V> m = new HashMap<K, V>(keys.size(), 1f);
        Iterator<K> k = keys.iterator();
        Iterator<V> v = values.iterator();
        while (k.hasNext() && v.hasNext())
        {
            m.put(k.next(), v.next());
        }
        return m;
    }

}
