/**
 *
 * Openutils testing library (http://www.openmindlab.com/products/testing.html)
 *
 * Copyright(C) 2005-2013, Openmind S.r.l. http://www.openmindonline.it
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * You may obtain a copy of the License at
 *
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package it.openutils.testing.collections;

import it.openutils.testing.AssertFormatter;

import java.util.Collection;
import java.util.Map;
import java.util.Set;


/**
 * @author gcatania
 */
public class MapAssert extends AssertFormatter
{

    protected MapAssert()
    {
    }

    /**
     * Asserts that a map is null
     * @param m the map
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertNull(Map< ? , ? > m, String message) throws AssertionError
    {
        if (m != null)
        {
            throw assertionFailed(message, "Expected a null {0}, found {1} values instead.", typeName(m), m.size());
        }
    }

    /**
     * Asserts that a map is null
     * @param m the map
     * @throws AssertionError if assertion fails
     */
    public static void assertNull(Map< ? , ? > m) throws AssertionError
    {
        assertNull(m, null);
    }

    /**
     * Asserts that a map is not null
     * @param m the map
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertNotNull(Map< ? , ? > m, String message) throws AssertionError
    {
        if (m == null)
        {
            throw assertionFailed(message, "{0} is null.", capitalizedTypeName(m));
        }
    }

    /**
     * Asserts that a map is not null
     * @param m the map
     * @throws AssertionError if assertion fails
     */
    public static void assertNotNull(Map< ? , ? > m) throws AssertionError
    {
        assertNotNull(m, null);
    }

    /**
     * Asserts that a map is empty. Fails when the collection is null.
     * @param m the map
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertEmpty(Map< ? , ? > m, String message) throws AssertionError
    {
        assertNotNull(m, message);
        if (!m.isEmpty())
        {
            throw assertionFailed(message, "Expected an empty {0}, found {1} values instead.", typeName(m), m.size());
        }
    }

    /**
     * Asserts that a map is empty. Fails when the collection is null.
     * @param m the map
     * @throws AssertionError if assertion fails
     */
    public static void assertEmpty(Map< ? , ? > m) throws AssertionError
    {
        assertEmpty(m, null);
    }

    /**
     * Asserts that a map has the expected size. Fails if the input collection is null.
     * @param actual the actual collection
     * @param expectedSize the expected size
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertSizeEquals(Map< ? , ? > actual, int expectedSize, String message) throws AssertionError
    {
        assertNotNull(actual, message);
        if (expectedSize == 0)
        {
            assertEmpty(actual, message);
        }
        else
        {
            int actualSize = actual.size();
            if (actualSize != expectedSize)
            {
                throw assertionFailed(
                    message,
                    "{0} has wrong size: expected {1}, found {2} instead.",
                    capitalizedTypeName(actual),
                    expectedSize,
                    actualSize);
            }
        }
    }

    /**
     * Asserts that a map has the expected size. Fails if the input collection is null.
     * @param actual the actual collection
     * @param expectedSize the expected size
     * @throws AssertionError if assertion fails
     */
    public static void assertSizeEquals(Map< ? , ? > actual, int expectedSize) throws AssertionError
    {
        assertSizeEquals(actual, expectedSize, null);
    }

    /**
     * Asserts that a map's key set equals an input key set
     * @param actual the actual map
     * @param expectedKeySet the expected key set
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertKeySetEquals(Map< ? , ? > actual, Set< ? > expectedKeySet, String message)
        throws AssertionError
    {
        assertNotNull(actual, message);
        SetAssert.assertEquals(actual.keySet(), expectedKeySet, message);
    }

    /**
     * Asserts that a map's key set equals an input key set
     * @param actual the actual map
     * @param expectedKeySet the expected key set
     * @throws AssertionError if assertion fails
     */
    public static void assertKeySetEquals(Map< ? , ? > actual, Set< ? > expectedKeySet) throws AssertionError
    {
        assertKeySetEquals(actual, expectedKeySet, null);
    }

    /**
     * Asserts that a map's value collection equals an input collection
     * @param actual the actual map
     * @param expectedValues the expected values
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertValuesEqual(Map< ? , ? > actual, Collection< ? > expectedValues, String message)
        throws AssertionError
    {
        assertNotNull(actual, message);
        CollectionAssert.assertEquals(actual.values(), expectedValues, message);
    }

    /**
     * Asserts that a map's value collection equals an input collection
     * @param actual the actual map
     * @param expectedValues the expected values
     * @throws AssertionError if assertion fails
     */
    public static void assertValuesEqual(Map< ? , ? > actual, Collection< ? > expectedValues) throws AssertionError
    {
        assertValuesEqual(actual, expectedValues, null);
    }

    /**
     * Asserts that the input map contains an entry with the input key and value
     * @param actual the actual map
     * @param expectedKey the key expected to be found in the map
     * @param expectedValue the value expected to be found for the given key
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertEntry(Map< ? , ? > actual, Object expectedKey, Object expectedValue, String message)
        throws AssertionError
    {
        assertNotNull(actual, message);
        if (!actual.containsKey(expectedKey))
        {
            throw assertionFailed(message, "Map does not contain an entry with key {0}.", expectedKey);
        }
        Object actualValue = actual.get(expectedKey);
        if (expectedValue == null)
        {
            if (actualValue != null)
            {
                throw assertionFailed(
                    message,
                    "Expected a null value mapped to key {0}, found {1} instead.",
                    expectedKey,
                    actualValue);
            }
        }
        else
        {
            if (!expectedValue.equals(actualValue))
            {
                throw assertionFailed(
                    message,
                    "Expected value {0} mapped to key {1}, found value {2} instead.",
                    expectedValue,
                    expectedKey,
                    actualValue);
            }
        }

    }

    /**
     * Asserts that the input map contains an entry with the input key and value
     * @param actual the actual map
     * @param expectedKey the key expected to be found in the map
     * @param expectedValue the value expected to be found for the given key
     * @throws AssertionError if assertion fails
     */
    public static void assertEntry(Map< ? , ? > actual, Object expectedKey, Object expectedValue) throws AssertionError
    {
        assertEntry(actual, expectedKey, expectedValue, null);
    }

    /**
     * Asserts that the two input maps are equal, regardless of entry order.
     * @param actual the actual map
     * @param expected the expected map
     * @param message a custom error message
     * @throws AssertionError if assertion fails
     */
    public static void assertEquals(Map< ? , ? > actual, Map< ? , ? > expected, String message) throws AssertionError
    {
        if (expected == null)
        {
            if (actual != null)
            {
                throw assertionFailed(message, "Expected a null map, found {0} entries instead.", actual.size());
            }
        }
        else if (expected.isEmpty())
        {
            if (!actual.isEmpty())
            {
                throw assertionFailed(message, "Expected an empty map, found {0} entries instead.", actual.size());
            }
        }
        else
        {
            assertKeySetEquals(actual, expected.keySet(), message);
            for (Map.Entry< ? , ? > e : expected.entrySet())
            {
                assertEntry(actual, e.getKey(), e.getValue(), message);
            }
        }
    }

    /**
     * Asserts that the two input maps are equal, regardless of entry order.
     * @param actual the actual map
     * @param expected the expected map
     * @throws AssertionError if assertion fails
     */
    public static void assertEquals(Map< ? , ? > actual, Map< ? , ? > expected) throws AssertionError
    {
        assertEquals(actual, expected, null);
    }

}
